//
//  AuthManager.swift
//  DIPSAuthPrototypeA
//
//  Created by trd-studs on 24/03/2021.
//

import AppAuth

class AuthManager : NSObject, OIDAuthStateChangeDelegate, URLSessionDelegate {
    static var shared = AuthManager()
    
    public var authState: OIDAuthState?
    private let authStateKey = "authState"
    
    public var customUrlSession: URLSession {
        return currentUrlSession ?? URLSession.shared
    }
    
    private var currentUrlSession: URLSession?
    
    override init() {
        super.init()
        self.loadState()
        #if DEBUG
            self.loadCustomURLSession()
        #endif
    }
    
    struct Constants {
        static var clientId: String = ""
        static var clientSecret: String = ""
        static var redirectURI: URL = URL(string: "dips-sso://login-callback")!
        static var scopes: [String] = ["openid", "offline_access", "profile"]
        static var authorizationEndpoint: URL = URL(string: "")!
        static var tokenEndpoint: URL = URL(string: "")!
    }
    
    var isSignedIn: Bool {
        return authState?.isAuthorized ?? false
    }
    
    func authorize(view: UIViewController, completion: (() -> Void)? = nil) {
        let configuration = getOIDCProviderConfiguration()
        self.authorizeWithAutoCodeExchange(view: view, configuration: configuration, clientId: Constants.clientId, clientSecret: Constants.clientSecret, redirectURI: Constants.redirectURI, scopes: Constants.scopes) {
            authState, error in
            if let authState = authState {
                self.setAuthState(authState)
                
                if let completion = completion {
                    completion()
                }
            } else {
                self.setAuthState(nil)
            }
            
            if let error = error {
                print("An error occured while fetching the authorization state: \(error.localizedDescription)")
            }
        }
    }
    
    private func getOIDCProviderConfiguration() -> OIDServiceConfiguration {
        let configuration = OIDServiceConfiguration.init(authorizationEndpoint: Constants.authorizationEndpoint, tokenEndpoint: Constants.tokenEndpoint)
        
        return configuration
    }
    
    private func loadCustomURLSession() {
        let urlConfiguration = URLSessionConfiguration.default
        let session = Foundation.URLSession(configuration: urlConfiguration, delegate: self, delegateQueue: OperationQueue.main)
        OIDURLSessionProvider.session()
        OIDURLSessionProvider.setSession(session)
        currentUrlSession = session
    }
    
    func didChange(_ state: OIDAuthState) {
        self.stateChanged()
    }
    
    func loadState() {
        guard let data = UserDefaults.standard.object(forKey: authStateKey) as? Data else {
            return
        }
        
        var authState: OIDAuthState? = nil
        
        if #available(iOS 12.0, *) {
            authState = try! NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(data) as? OIDAuthState
        } else {
            authState = NSKeyedUnarchiver.unarchiveObject(with: data) as? OIDAuthState
        }
        
        if let authState = authState {
            self.setAuthState(authState)
        }
    }
    
    private func stateChanged() {
        self.saveState()
    }
    
    private func saveState() {
        var data: Data? = nil
        
        if let authState = self.authState {
            if #available(iOS 12.0, *) {
                data = try! NSKeyedArchiver.archivedData(withRootObject: authState, requiringSecureCoding: false)
            } else {
                data = NSKeyedArchiver.archivedData(withRootObject: authState)
            }
        }
        
        UserDefaults.standard.set(data, forKey: authStateKey)
        UserDefaults.standard.synchronize()
    }
    
    private func setAuthState(_ authState: OIDAuthState?) {
        if self.authState != authState {
            self.authState = authState
            self.authState?.stateChangeDelegate = self
            self.stateChanged()
        }
    }
    
    private func authorizeWithAutoCodeExchange(view: UIViewController,
                                        configuration: OIDServiceConfiguration,
                                       clientId: String,
                                       clientSecret: String,
                                       redirectURI: URL,
                                       scopes: [String] = [OIDScopeOpenID, OIDScopeProfile],
                                       completion: @escaping (OIDAuthState?, Error?) -> Void) {
        guard let appDelegate = UIApplication.shared.delegate as? AppDelegate else {
            print("Error accessing AppDelegate")
            return
        }
        
        let request = OIDAuthorizationRequest(configuration: configuration,
                                              clientId: clientId,
                                              clientSecret: clientSecret,
                                              scopes: scopes,
                                              redirectURL: redirectURI,
                                              responseType: OIDResponseTypeCode,
                                              additionalParameters: nil)
        
        appDelegate.currentAuthorizationFlow = OIDAuthState.authState(byPresenting: request, presenting: view) { authState, error in
            completion(authState, error)
        }
    }
    
    func urlSession(_ session: URLSession, didReceive challenge: URLAuthenticationChallenge, completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        completionHandler(Foundation.URLSession.AuthChallengeDisposition.useCredential, URLCredential(trust: challenge.protectionSpace.serverTrust!))
    }
}
