//
//  AuthManager.swift
//  SSOExtension
//
//  Created by trd-studs on 26/04/2021.
//

import Foundation
import AppAuth

// This data should be deployed through MDM
struct Constants {
    static var authorizationEndpoint: URL = URL(string: "https://vt-mob-dev2.dips.local/dips.oauth/identity/connect/authorize")!
    static var tokenEndpoint: URL = URL(string: "https://vt-mob-dev2.dips.local/dips.oauth/identity/connect/token")!
    static var scopes: [String] = ["openid", "offline_access", "profile"]
}

class AuthManager : NSObject, OIDAuthStateChangeDelegate, URLSessionDelegate {
    static var shared = AuthManager()
    
    private var preferences: Preferences
    
    override init() {
        preferences = Utils.getPreferences()
    }
    
    public var authState: OIDAuthState?
    private let authStateKey = "authState"
    
    private var currentAuthorizationFlow : OIDExternalUserAgentSession?
    
    public var customUrlSession: URLSession {
        return currentUrlSession ?? URLSession.shared
    }
    
    private var currentUrlSession: URLSession?
    
    public func authorize(url: URL, view: UIViewController, completion: (() -> Void)? = nil) {
        let configuration = getOIDCProviderConfiguration()
        let authParameters = Utils.getAuthParamsFromURL(url: url)
        self.authorizeWithCodeExchange(view: view, configuration: configuration, clientId: preferences.clientId, clientSecret: preferences.clientSecret, redirectURI: preferences.redirectURI, scopes: preferences.scopes) {
            authState, error in
            if let authState = authState {
                self.setAuthState(authState)
                
                if let completion = completion {
                    completion()
                }
            } else {
                self.setAuthState(nil)
            }
            
            if let error = error {
                print("An error occured while fetching the authorization: \(error.localizedDescription)")
            }
        }
    }
    
    private func authorizeWithCodeExchange(view: UIViewController,
                                           configuration: OIDServiceConfiguration,
                                           clientId: String,
                                           clientSecret: String,
                                           redirectURI: URL,
                                           scopes: [String] = [OIDScopeOpenID, OIDScopeProfile],
                                           completion: @escaping (OIDAuthState?, Error?) -> Void) {
        
        let request = OIDAuthorizationRequest(configuration: configuration,
                                              clientId: clientId,
                                              clientSecret: clientSecret,
                                              scopes: scopes,
                                              redirectURL: redirectURI,
                                              responseType: OIDResponseTypeCode,
                                              additionalParameters: nil)
        
        currentAuthorizationFlow = OIDAuthState.authState(byPresenting: request, presenting: view) { authState, error in
            completion(authState, error)
        }
    }
    
    func didChange(_ state: OIDAuthState) {
        self.stateChanged()
    }
    
    private func getOIDCProviderConfiguration() -> OIDServiceConfiguration {
        let configuraion = OIDServiceConfiguration.init(authorizationEndpoint: preferences.authorizationEndpoint, tokenEndpoint: preferences.tokenEndpoint)
        
        return configuraion
    }
    
    // This function sets up URLSession to allow all certificates
    private func loadCustomURLSession() {
        let urlConfiguration = URLSessionConfiguration.default
        let session = Foundation.URLSession(configuration: urlConfiguration, delegate: self, delegateQueue: OperationQueue.main)
        OIDURLSessionProvider.session()
        OIDURLSessionProvider.setSession(session)
        currentUrlSession = session
    }
    
    func urlSession(_ session: URLSession, didReceive challenge: URLAuthenticationChallenge, completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        completionHandler(Foundation.URLSession.AuthChallengeDisposition.useCredential, URLCredential(trust: challenge.protectionSpace.serverTrust!))
    }
    
    private func loadState() {
        guard let data = UserDefaults.standard.object(forKey: authStateKey) as? Data else {
            return
        }
        
        var authState: OIDAuthState? = nil
        
        if #available(iOS 12.0, *) {
            authState = try! NSKeyedUnarchiver.unarchiveTopLevelObjectWithData(data) as? OIDAuthState
        } else {
            authState = NSKeyedUnarchiver.unarchiveObject(with: data) as? OIDAuthState
        }
        
        if let authState = authState {
            self.setAuthState(authState)
        }
    }
    
    private func stateChanged() {
        self.saveState()
    }
    
    private func saveState() {
        var data: Data? = nil
        if let authState = self.authState {
            if #available(iOS 12.0, *) {
                data = try! NSKeyedArchiver.archivedData(withRootObject: authState, requiringSecureCoding: false)
            } else {
                data = NSKeyedArchiver.archivedData(withRootObject: authState)
            }
        }
        
        UserDefaults.standard.set(data, forKey: authStateKey)
        UserDefaults.standard.synchronize()
    }
    
    private func setAuthState(_ authState: OIDAuthState?) {
        if self.authState != authState {
            self.authState = authState
            self.authState?.stateChangeDelegate = self
            self.stateChanged()
        }
    }
}
